@@ -333,6 +333,22 @@ sub lcm {
   _validate_integer($_) for @v;
   return Math::Prime::Util::PP::lcm(@v);
 }
+sub gcdext {
+  my($a,$b) = @_;
+  _validate_integer($a);
+  _validate_integer($b);
+  return Math::Prime::Util::PP::gcdext($a,$b);
+}
+sub chinese {
+  # TODO: make sure we're not modding their data
+  foreach my $aref (@_) {
+    die "chinese arguments are two-element array references"
+      unless ref($aref) eq 'ARRAY' && scalar @$aref == 2;
+    _validate_integer($aref->[0]);
+    _validate_integer($aref->[1]);
+  }
+  return Math::Prime::Util::PP::chinese(@_);
+}
 sub vecsum {
   my(@v) = @_;
   _validate_integer($_) for @v;
