@@ -43,16 +43,23 @@ ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/
 #if USE_MONT_PRIMALITY
+
+#if defined(__GNUC__) && (__GNUC__ >= 3)
+ #define MPU_UNLIKELY(x) __builtin_expect(!!(x), 0)
+#else
+ #define MPU_UNLIKELY(x) (x)
+#endif
+
 static INLINE uint64_t mont_prod64(uint64_t a, uint64_t b, uint64_t n, uint64_t npi)
 {
   uint64_t t_hi, t_lo, m, mn_hi, mn_lo, u;
   /* t_hi * 2^64 + t_lo = a*b */
   asm("mulq %3" : "=a"(t_lo), "=d"(t_hi) : "a"(a), "rm"(b));
+  if (MPU_UNLIKELY(t_lo == 0)) return t_hi;
   m = t_lo * npi;
   /* mn_hi * 2^64 + mn_lo = m*n */
   asm("mulq %3" : "=a"(mn_lo), "=d"(mn_hi) : "a"(m), "rm"(n));
-  u = t_hi + mn_hi;
-  if (t_lo + mn_lo < t_lo) u++;
+  u = t_hi + mn_hi + 1;
   return (u < t_hi || u >= n)  ?  u-n  :  u;
 }
 #define mont_square64(a, n, npi)  mont_prod64(a, a, n, npi)
@@ -70,39 +77,26 @@ static INLINE UV mont_powmod64(uint64_t a, uint64_t k, uint64_t one, uint64_t n,
  * Divisionless Operations", Computer Journal (1994) 37 (3): 219-222, Proc 5 */
 static INLINE uint64_t modular_inverse64(const uint64_t a)
 {
-#if 0
   uint64_t S = 1, J = 0;
-  int i;
-  for (i = 0; i < 64; i++) {
-    if (S & 1) {
-      J |= (1ULL << i);
-      S += a;
-    }
-    S >>= 1;
-  }
-#else
-  /* 2 at a time. */
-  uint64_t S = 1, J = 0;
-  const uint64_t a1p = (a>>1)+1, a2 = a>>2, a2p = a2+1;
-  int i;
-  if (a & 2) {
-    const uint64_t tab[4] = { 0,  a2p,  a1p,  a1p+a2p };
-    for (i = 0; i < 32; i++) {
-      uint64_t S2 = S >> 2;
-      int index = S & 3;
-      J |= (uint64_t)index << (2 * i);
-      S = S2 + tab[index];
-    }
-  } else {
-    const uint64_t tab[4] = { 0,  a1p+a2,  a1p,  a2p };
-    for (i = 0; i < 32; i++) {
-      uint64_t S2 = S >> 2;
-      int index = S & 3;
-      J |= (uint64_t)((4-index)&3) << (2 * i);
-      S = S2 + tab[index];
-    }
-  }
-#endif
+  int idx;
+  /* Basic algorithm:
+   *    for (i = 0; i < 64; i++) {
+   *      if (S & 1)  {  J |= (1ULL << i);  S += a;  }
+   *      S >>= 1;
+   *    }
+   * What follows is 8 bits at a time, unrolled by hand. */
+  static const char mask[128] = {255,85,51,73,199,93,59,17,15,229,195,89,215,237,203,33,31,117,83,105,231,125,91,49,47,5,227,121,247,13,235,65,63,149,115,137,7,157,123,81,79,37,3,153,23,45,11,97,95,181,147,169,39,189,155,113,111,69,35,185,55,77,43,129,127,213,179,201,71,221,187,145,143,101,67,217,87,109,75,161,159,245,211,233,103,253,219,177,175,133,99,249,119,141,107,193,191,21,243,9,135,29,251,209,207,165,131,25,151,173,139,225,223,53,19,41,167,61,27,241,239,197,163,57,183,205,171,1};
+
+  const char amask = mask[(a >> 1) & 127];
+  uint32_t T;
+  idx = (amask*(S&255)) & 255;  J = idx;                  S = (S+a*idx) >> 8;
+  idx = (amask*(S&255)) & 255;  J |= (uint64_t)idx << 8;  S = (S+a*idx) >> 8;
+  idx = (amask*(S&255)) & 255;  J |= (uint64_t)idx <<16;  S = (S+a*idx) >> 8;
+  idx = (amask*(S&255)) & 255;  J |= (uint64_t)idx <<24;  T = (S+a*idx) >> 8;
+  idx = (amask*(T&255)) & 255;  J |= (uint64_t)idx <<32;  T = (T+a*idx) >> 8;
+  idx = (amask*(T&255)) & 255;  J |= (uint64_t)idx <<40;  T = (T+a*idx) >> 8;
+  idx = (amask*(T&255)) & 255;  J |= (uint64_t)idx <<48;  T = (T+a*idx) >> 8;
+  idx = (amask*(T&255)) & 255;  J |= (uint64_t)idx <<56;
   return J;
 }
 static INLINE uint64_t compute_modn64(const uint64_t n)
