@@ -45,8 +45,10 @@ extern int kronecker_su(IV a, UV b);
 extern int kronecker_ss(IV a, IV b);
 
 extern UV binomial(UV n, UV k);
-extern UV modinverse(UV a, UV p);    /* Returns 1/a mod p */
-extern UV divmod(UV a, UV b, UV n);  /* Returns a/b mod n */
+extern IV gcdext(IV a, IV b, IV* u, IV* v, IV* s, IV* t); /* Ext Euclidean */
+extern UV modinverse(UV a, UV p);              /* Returns 1/a mod p */
+extern UV divmod(UV a, UV b, UV n);            /* Returns a/b mod n */
+extern UV chinese(UV* a, UV* n, UV num, int *status); /* Chinese Remainder */
 
 extern UV totient(UV n);
 extern int moebius(UV n);
